/* 
 ** Copyright (C) 2010 by Alessandro Galetto
 **
 ** VERSION
 **+-------------------------------------------------------------------------*
 ** 1.0.0
 **   
 ** MODULE AUTHOR(S):
 **+-------------------------------------------------------------------------*
 **  ag
 **
 ** MODULE TITLE:
 **+-------------------------------------------------------------------------*
 ** main.c
 **
 ** MODULE FUNCTION:
 **+-------------------------------------------------------------------------*
 ** This is a basic command line utility for Mac OS X that will set the Led 
 ** mini panel display scrolling text.
 ** 
 **
 ** MODIFICATION HISTORY
 **
 ** Date      Author         Comments
 **+---------+--------------+------------------------------------------------*
 ** 20100512  ag             Created this file
 **+---------+--------------+------------------------------------------------*
 **
 ** LICENSE
 **
 ****************************************************************************
 **   This program is free software; you can redistribute it and/or modify  *
 **   it under the terms of the GNU General Public License as published by  *
 **   the Free Software Foundation; either version 2 of the License, or     *
 **   (at your option) any later version.                                   *
 **                                                                         *
 **   This program is distributed in the hope that it will be useful,       *
 **   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 **   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 **   GNU General Public License for more details.                          *
 **                                                                         *
 **   You should have received a copy of the GNU General Public License     *
 **   along with this program; if not, write to the                         *
 **   Free Software Foundation, Inc.,                                       *
 **   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ****************************************************************************
 */

/*
** INCLUDES
*/
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/serial/IOSerialKeys.h>
#if defined(MAC_OS_X_VERSION_10_3) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_3)
#include <IOKit/serial/ioss.h>
#endif
#include <IOKit/IOBSD.h>
#include <AvailabilityMacros.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <paths.h>
#include <termios.h>
#include <sysexits.h>
#include <sys/param.h>
#include <sys/select.h>
#include <sys/time.h>
#include <time.h>

/*
** DEFINES
*/
#define DEBUG 0


/*
** GLOBALS
*/
char programName[]	=	"mlpupdate";
char mlpInit[]		=	{0x54, 0x41, 0x42, 0x43, 0x44, 0x45, 0x00};
char mlpEnd[]		=	{0x02, 0x33, 0x1};

/*
 *****************************************************************************
 **
 ** FUNCTION AUTHOR(S):
 ** 	ag
 **
 ** FUNCTION NAME: 
 ** 	msleep
 **
 ** FUNCTION TYPE
 **		void
 **
 ** FUNCTION ARGUMENTS: 
 ** 	unsigned long millisec - Number of milliseconds to sleep
 **
 ** FUNCTION RETURNS:
 ** 	N/A
 **
 ** FUNCTION DESCRIPTION
 ** 	Sleep for the number of milliseconds passed as an argument. 
 **
 *****************************************************************************
 */
void msleep(unsigned long millisec)
{
    struct timespec req={0};
    time_t sec=(int)(millisec/1000);
    millisec=millisec-(sec*1000);
    req.tv_sec=sec;
    req.tv_nsec=millisec*1000000L;
    while(nanosleep(&req,&req)==-1)
		continue;
}


/*
 *****************************************************************************
 **
 ** FUNCTION AUTHOR(S):
 ** 	ag
 **
 ** FUNCTION NAME: 
 ** 	OpenSerialPort
 **
 ** FUNCTION TYPE
 **		int
 **
 ** FUNCTION ARGUMENTS: 
 ** 	const char *bsdPath - Serial device BSD name
 **		struct termios *gOriginalTTYAttrs - termios structure to save TTY original settings
 **
 ** FUNCTION RETURNS:
 ** 	-1	Errors have been encountered opening the serial port or configuring it
 **		int	file description of the serial port
 **  
 **
 ** FUNCTION DESCRIPTION
 ** 	This function increments the value of twitter_conversation_id 
 **
 *****************************************************************************
 */ 
static int OpenSerialPort(const char *bsdPath, struct termios *gOriginalTTYAttrs)
{
    int				fileDescriptor = -1;
    struct			termios	options;
        
    fileDescriptor = open(bsdPath, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if (fileDescriptor == -1) 
	{
        printf("%s: Error opening serial port %s - %s(%d).\n",
               programName, bsdPath, strerror(errno), errno);
        goto error;
    }
	
		
    // Get the current options and save them so we can restore the default settings later.
    if (tcgetattr(fileDescriptor, gOriginalTTYAttrs) == -1) 
	{
        printf("%s: Error getting tty attributes %s - %s(%d).\n",
			   programName, bsdPath, strerror(errno), errno);
        goto error;
    }
	
    // The serial port attributes such as timeouts and baud rate are set by modifying the termios
    // structure and then calling tcsetattr() to cause the changes to take effect. Note that the
    // changes will not become effective without the tcsetattr() call.
    // See tcsetattr(4) ("man 4 tcsetattr") for details.
    
    options = *gOriginalTTYAttrs;
    
    // Print the current input and output baud rates.
    // See tcsetattr(4) ("man 4 tcsetattr") for details.
    
    //printf("Current input baud rate is %d\n", (int) cfgetispeed(&options));
    //printf("Current output baud rate is %d\n", (int) cfgetospeed(&options));
    
    // Set raw input (non-canonical) mode, with reads blocking until either a single character 
    // has been received or a one second timeout expires.
    // See tcsetattr(4) ("man 4 tcsetattr") and termios(4) ("man 4 termios") for details.
    
    //cfmakeraw(&options);
    options.c_cc[VMIN] = 1;
    options.c_cc[VTIME] = 10;
	
    // Setting cflags options
    cfsetspeed(&options, B38400);
    options.c_cflag = 0;
	options.c_cflag |= (CREAD | CS8 | HUPCL | CLOCAL);
	
#if defined(MAC_OS_X_VERSION_10_4) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_4)
	speed_t speed = 38400; 
    if (ioctl(fileDescriptor, IOSSIOSPEED, &speed) == -1) 
	{
        printf("%s: Error calling ioctl(..., IOSSIOSPEED, ...) %s - %s(%d).\n",
			   programName, bsdPath, strerror(errno), errno);
    }
#endif
        
    // Cause the new options to take effect immediately.
    if (tcsetattr(fileDescriptor, TCSANOW, &options) == -1) 
	{
        printf("%s: Error setting tty attributes %s - %s(%d).\n",
			   programName, bsdPath, strerror(errno), errno);
        goto error;
    }
	
	// Assert Data Terminal Ready (DTR)
    if (ioctl(fileDescriptor, TIOCSDTR) == -1)  
	{
        printf("%s: Error asserting DTR %s - %s(%d).\n",
			   programName, bsdPath, strerror(errno), errno);
    }
    
	// Clear Data Terminal Ready (DTR)
    if (ioctl(fileDescriptor, TIOCCDTR) == -1)  
	{
        printf("%s: Error clearing DTR %s - %s(%d).\n",
			   programName, bsdPath, strerror(errno), errno);
    }
    
    
    // Success
    return fileDescriptor;
    
    // Failure path
error:
    if (fileDescriptor != -1)
    {
        close(fileDescriptor);
    }
    
    return -1;
}


/*
 *****************************************************************************
 **
 ** FUNCTION AUTHOR(S):
 ** 	ag
 **
 ** FUNCTION NAME: 
 ** 	CloseSerialPort
 **
 ** FUNCTION TYPE
 **		void
 **
 ** FUNCTION ARGUMENTS: 
 ** 	const char *bsdPath - Serial device BSD name
 **		struct termios *gOriginalTTYAttrs - termios structure to restore TTY original settings
 **
 ** FUNCTION RETURNS:
 ** 	N/A
 **
 ** FUNCTION DESCRIPTION
 ** 	This function simply waits for all output to be sent to device, restores
 **     TTY settings and closes the serial port
 **
 *****************************************************************************
 */ 
void CloseSerialPort(int fileDescriptor, struct termios *gOriginalTTYAttrs)
{
    // Block until all written output has been sent from the device.
    // Note that this call is simply passed on to the serial device driver. 
	// See tcsendbreak(3) ("man 3 tcsendbreak") for details.
    if (tcdrain(fileDescriptor) == -1) 
	{
        printf("%s: Error waiting for drain - %s(%d).\n",
			   programName, strerror(errno), errno);
    }
	
	// Actually it seems that the PL2303 driver on Mac OS X does not implement
	// any drain function so we will just wait for one second to be sure that
	// all data is sent to the device. Yeah, it's not that safe.
	msleep(1000);
    
    if (tcsetattr(fileDescriptor, TCSANOW, gOriginalTTYAttrs) == -1) 
	{
        printf("%s: Error resetting tty attributes - %s(%d).\n",
			   programName, strerror(errno), errno);
    }
	
    close(fileDescriptor);
}


/*
 *****************************************************************************
 **
 ** FUNCTION AUTHOR(S):
 ** 	ag
 **
 ** FUNCTION NAME: 
 ** 	sendData
 **
 ** FUNCTION TYPE
 **		int
 **
 ** FUNCTION ARGUMENTS: 
 ** 	int fileDescriptor - This is the file descriptor where to send data
 **		char * data - Data to be sent to the device
 **
 ** FUNCTION RETURNS:
 ** 	-1 Error while sending data to the device
 **     0  No errors.
 **
 ** FUNCTION DESCRIPTION
 ** 	This function sends data to the device according to its protocol
 **
 *****************************************************************************
 */
int sendData(int fileDescriptor, char *data) 
{
	
	ssize_t				numBytes;
	unsigned char		packet[69];
	int					numPackets;
	int					i, l, m, length, checksum;
	int					sequence;
	
	// Send initialization data
	// The initialization data is made out of seven bytes 0x54, 0x41, 0x42, 0x43, 0x44, 0x45, 0x00
	// As of today I have not been able to understand the meaning of these seven bytes.
	numBytes = write(fileDescriptor, mlpInit, sizeof(mlpInit));
	if (numBytes == -1)
	{
		printf("%s: Error writing to device - %s(%d).\n", programName, strerror(errno), errno);
		return(-1);
	}
	
	numPackets = strlen(data) / 60;
	if ((strlen(data) % 60) != 0)
		numPackets++;
	
	for (i=0; i<numPackets; i++) {
		memset(packet, 0x00, sizeof(packet));
		if (DEBUG)
			printf("Packet number %d\n", i+1);
		packet[0] = 0x02;				// Start of packet
		packet[1] = 0x31;				// Packet type - TBC
		packet[2] = 0x06;				// Line Number
		sequence = 0x00 + 0x40 *i;
		packet[3] = sequence;	// Sequence number
		if (DEBUG)
			printf("\t Sequence: %x\n", sequence);
		packet[4] = 0x35;				// Scroll speed
		packet[5] = 0x31;
		packet[6] = 0x42;
		packet[7] = 0x01;
		
		// Copy the data into the packet
		length = strlen(data+i*60);
		if (length > 60)
			length = 60;
		strncpy((char *)&packet[8], data+i*60, length);
		if (DEBUG)
			printf("\t Length: %d\n", length);
		
		if (DEBUG)
			printf("\t Hex Dump\n");
		for (l=0; l<68; l++) {
			if (DEBUG)
				printf("%x ", packet[l]);
		}
		if (DEBUG)
			printf("\n");
		
		// Calculate checksum
		checksum = 0x00;
		for (l=0; l<67; l++) {
			checksum += packet[l+1];
		}
		if (DEBUG)
			printf("\t Checksum count: %x\n", checksum);
		checksum = checksum % 0x100;
		if (DEBUG)
			printf("\t Checksum: %x\n", checksum);
		packet[68] = checksum;
		
		numBytes = write(fileDescriptor, packet, sizeof(packet));
		if (numBytes == -1)
		{
			printf("%s: Error writing to device - %s(%d).\n", programName, strerror(errno), errno);
			return(-1);
		}
	}
	
	m = 4 - numPackets;
	for (i=0; i<m; i++) {
		memset(packet, 0x00, sizeof(packet));
		if (DEBUG)
			printf("Packet number %d\n", numPackets+i+1);
		packet[0] = 0x02;							// Start of packet
		packet[1] = 0x31;							// Packet type - TBC
		packet[2] = 0x06;							// Line Number
		sequence = 0x00 + (0x40 * (i+numPackets));
		if (DEBUG)
			printf("\t Sequence index: %d\n", sequence);
		packet[3] = sequence;	// Sequence number
		if (DEBUG)
			printf("\t Sequence: %x\n", sequence);
		
		if (DEBUG)
			printf("\t Hex Dump\n");
		for (l=0; l<68; l++) {
			if (DEBUG)
				printf("%x ", packet[l]);
		}
		if (DEBUG)
			printf("\n");
		
		// Calculate checksum
		checksum = 0x00;
		for (l=0; l<67; l++) {
			checksum += packet[l+1];
		}
		if (DEBUG)
			printf("\t Checksum count: %x\n", checksum);
		checksum = checksum % 0x100;
		if (DEBUG)
			printf("\t Checksum: %x\n", checksum);
		packet[68] = checksum;
		
		numBytes = write(fileDescriptor, packet, sizeof(packet));
		if (numBytes == -1)
		{
			printf("%s: Error writing to device - %s(%d).\n", programName, strerror(errno), errno);
			return(-1);
		}
	}

	
	// Send end of communication data
	// The end of communication data is made out of three bytes 0x02, 0x33, 0x1
	// As of today I have not been able to understand the meaning of these three bytes.
	numBytes = write(fileDescriptor, mlpEnd, sizeof(mlpEnd));
	if (numBytes == -1)
	{
		printf("%s: Error writing to device - %s(%d).\n", programName, strerror(errno), errno);
		return(-1);
	}
	
	
	return(0);
}


int main (int argc, const char * argv[]) {
	
	char					bsdPath[MAXPATHLEN] = "/dev/cu.usbserial";
	char					data[MAXPATHLEN];
	int						fileDescriptor;
	struct termios			gOriginalTTYAttrs;
	int						i, count;
	
	memset(&gOriginalTTYAttrs, 0x00, sizeof(struct termios));
	fileDescriptor = OpenSerialPort(bsdPath, &gOriginalTTYAttrs);
    if (-1 == fileDescriptor)
    {
        return EX_IOERR;
    }
	
	if (argc > 1) {
		if (DEBUG)
			printf("Data to be sent: <%s>\n", argv[1]);
		memset(data, 0x00, sizeof(data));
		count = 0;
		for (i=0; i<strlen(argv[1])+1; i++) {
			if (argv[1][i] != '"')
				data[count++] = argv[1][i];
		}
		if (strlen(data) > 240)
			data[240] = 0x00;
		
		sendData(fileDescriptor, data);
	} 
	
    CloseSerialPort(fileDescriptor, &gOriginalTTYAttrs);
		
    return EX_OK;
}
